@@ -21,6 +21,8 @@ Where verb is one of:
  * list-pinned-tags: Idem, but only pinned tags.
  * list-packages: List all packages whose name contains <argument>.
  * list-installed-packages: Idem, but only installed packages.
+ * uninstall-unpinned: Keep only the pinned tags and all dependencies.
+ * collect-pool: Remove pool files not needed by any installed package.
 
 Examples:
 %(progname)s pin xta:latest           # installs latest XTA
@@ -30,19 +32,8 @@ Examples:
 	sys.exit(1)
 
 
-# Read configuration (i.e. list of pinned tags)
-config_path = os.path.join(rapid.content_dir, 'main.cfg')
-config = ConfigParser()
-config.read(config_path)
-pinned_tags = []
-
-if not config.has_section('tags'):
-	config.add_section('tags')
-if config.has_option('tags', 'pinned'):
-	pinned_tags = set([s for s in config.get('tags', 'pinned').split(',') if s])
-
-
 #  Create rapid module.
+pool_dir = rapid.pool_dir
 rapid = rapid.Rapid()
 
 
@@ -77,33 +68,33 @@ def select(noun, needle, haystack):
 def pin_single(tag):
 	""" Pin a tag. This means any package having this tag will automatically be
 	    installed and upgraded."""
-	if not tag in pinned_tags:
+	if not tag in rapid.pinned_tags():
 		print 'Pinning: ' + tag
-		pinned_tags.add(tag)
+		rapid.pinned_tags().add(tag)
 	else:
 		print 'Already pinned: ' + tag
 
 
 def pin(searchterm):
 	""" Pin all tags matching searchterm and install the corresponding packages."""
-	for t in select('tag', searchterm, rapid.get_packages_by_tag().iterkeys()):
+	for t in select('tag', searchterm, rapid.tags()):
 		pin_single(t)
-		install_single(rapid.get_package_by_tag(t))
+		install_single(rapid.packages()[t])
 
 
 def unpin_single(tag):
 	""" Unpin a tag. This means packages having this tag will not be
 	    automatically upgraded anymore. Does not uninstall anything."""
-	if tag in pinned_tags:
+	if tag in rapid.pinned_tags():
 		print 'Unpinning: ' + tag
-		pinned_tags.remove(tag)
+		rapid.pinned_tags().remove(tag)
 	else:
 		print 'Not pinned: ' + tag
 
 
 def unpin(searchterm):
 	""" Unpin all tags matching searchterm."""
-	for t in select('pinned tag', searchterm, pinned_tags):
+	for t in select('pinned tag', searchterm, rapid.pinned_tags()):
 		unpin_single(t)
 
 
@@ -122,8 +113,8 @@ def install_single(p, dep = False):
 
 def install(searchterm):
 	""" Install all packages matching searchterm."""
-	for name in select('name', searchterm, rapid.get_names().iterkeys()):
-		install_single(rapid.get_package_by_name(name))
+	for name in select('name', searchterm, [p.name for p in rapid.packages()]):
+		install_single(rapid.packages()[name])
 
 
 def uninstall_single(p):
@@ -138,21 +129,35 @@ def uninstall_single(p):
 		p.uninstall()
 
 
+def uninstall_single_plus_revdeps(p, dep = False):
+	""" Uninstall and unpin a package and all packages that depend on it."""
+	if p:
+		for d in p.reverse_dependencies:
+			uninstall_single_plus_revdeps(d, True)
+		for t in p.tags:
+			unpin_single(t)
+		if p.installed():
+			print ['Uninstalling: ', 'Uninstalling dependency: '][int(dep)] + p.name
+			p.uninstall()
+		elif not dep:
+			print 'Already uninstalled: ' + p.name
+
+
 def uninstall(searchterm):
 	""" Uninstall all packages matching searchterm."""
 	for name in select('name', searchterm, [p.name for p in rapid.get_installed_packages()]):
-		uninstall_single(rapid.get_package_by_name(name))
+		uninstall_single(rapid.packages()[name])
 
 
 def list_packages(searchterm, available):
 	""" List all packages whose name matches searchterm."""
 	s = searchterm.lower()
 	print 'Installed packages:'
-	for p in filter(lambda p: s in p.name.lower(), rapid.get_installed_packages()):
+	for p in [p for p in rapid.packages() if p.installed() and s in p.name.lower()]:
 		print '  %-40s (%s)' % (p.name, ', '.join(p.tags))
 	if available:
 		print 'Available packages:'
-		for p in rapid.get_not_installed_packages():
+		for p in [p for p in rapid.packages() if not p.installed() and s in p.name.lower()]:
 			print '  %-40s (%s)' % (p.name, ', '.join(p.tags))
 
 
@@ -160,23 +165,88 @@ def list_tags(searchterm, available):
 	""" List all tags which match searchterm."""
 	s = searchterm.lower()
 	print 'Pinned tags:'
-	for tag in filter(lambda t: s in t.lower(), pinned_tags):
-		p = rapid.get_package_by_tag(tag)
+	for tag in filter(lambda t: s in t.lower(), rapid.pinned_tags()):
+		p = rapid.packages()[tag]
 		if p:
 			print '  %-40s (%s)' % (tag, p.name)
 		else:
 			print '  %-40s [dangling tag]' % tag
 	if available:
 		print 'Available tags:'
-		for tag in (set(rapid.get_packages_by_tag()) - pinned_tags):
-			p = rapid.get_package_by_tag(tag)
+		for tag in (set(rapid.tags()) - set(rapid.pinned_tags())):
+			p = rapid.packages()[tag]
 			print '  %-40s (%s)' % (tag, p.name)
 
 
 def upgrade(searchterm):
 	""" Upgrade installed tags which match searchterm."""
-	for tag in filter(lambda t: searchterm.lower() in t.lower(), pinned_tags):
-		install_single(rapid.get_package_by_tag(tag))
+	for tag in filter(lambda t: searchterm.lower() in t.lower(), rapid.pinned_tags()):
+		install_single(rapid.packages()[tag])
+
+
+def uninstall_unpinned():
+	""" Simple mark and sweep garbage collector. The root set consists of the
+	    pinned tags. This does not touch pool files."""
+	# Build set marked_packages that should be kept.
+	marked_packages = set()
+	new_packages = set([rapid.packages()[t] for t in rapid.pinned_tags() if t in rapid.tags()])
+	while new_packages:
+		marked_packages.update(new_packages)
+		new_packages = sum([list(p.dependencies) for p in new_packages], [])
+
+	# Build set of packages that will be removed.
+	garbage = set(set([p for p in rapid.packages() if p.installed()]) - marked_packages)
+
+	# Confirmation?
+	if (not raw_input('Uninstall %s? [y/N]: ' % ', '.join([p.name for p in garbage])).startswith('y')):
+		return
+
+	# Uninstall all garbage.
+	for p in garbage:
+		if p.installed():
+			uninstall_single_plus_revdeps(p)
+
+
+def collect_pool():
+	""" Simple mark and sweep garbage collector. The root set consists of the
+	    installed packages. This touches only pool files."""
+	# Build set marked_files that should be kept.
+	installed_packages = [p for p in rapid.packages() if p.installed()]
+	marked_files = reduce(lambda x, y: x + y.get_files(), installed_packages, [])
+	marked_files = set([f.get_pool_path() for f in marked_files])
+
+	def gc(really_remove):
+		# Remove all files not in marked_files.
+		count = 0
+		size = 0
+		for i in range(0, 256):
+			d = os.path.join(pool_dir, '%02x' % i)
+			for f in os.listdir(d):
+				f = os.path.join(d, f)
+				if not f in marked_files:
+					count += 1
+					size += os.path.getsize(f)
+					if really_remove: os.unlink(f)
+		return (count, size)
+
+	count, size = gc(False)
+
+	print '%.2f megabytes / %d files can be collected from the pool.' % (size / (1024.*1024.), count)
+	if count == 0:
+		return
+
+	# Confirmation?
+	print
+	print '#' * 80
+	print '# %-76s #' % 'Collecting pool files will delete the data from disk permantly.'.center(76)
+	print '# %-76s #' % 'You will have to download the data again in case you need it.'.center(76)
+	print '#' * 80
+	print
+	if (not raw_input('Delete %d pool files? [y/N]: ' % count).startswith('y')):
+		return
+
+	count, size = gc(True)
+	print '%.2f megabytes / %d files deleted from the pool.' % (size / (1024.*1024.), count)
 
 
 def req_arg():
